﻿using Microsoft.Win32.SafeHandles;
using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Runtime.InteropServices;
using System.Security.Principal;
using System.Text;

namespace NtApi
{
  public static class NtSystemCalls
  {
    [DllImport("ntdll.dll")]
    public static extern int RtlNtStatusToDosError(int status);

    public static void StatusToNtException(int status)
    {
      if (status < 0)
      {
        throw new NtException(status);
      }
    }

    public static void StatusToWin32Exception(int status)
    {
      if (status < 0)
      {
        throw new Win32Exception(RtlNtStatusToDosError(status));
      }
    }

    [DllImport("ntdll.dll")]
    public static extern int NtOpenFile(
        out IntPtr FileHandle,
        FileAccessRights DesiredAccess,
        ObjectAttributes ObjAttr,
        [In] [Out] IoStatus IoStatusBlock,
        ShareMode ShareAccess,
        FileOpenOptions OpenOptions);

    public static SafeFileHandle OpenFile(string name, FileAccessRights DesiredAccess, ShareMode ShareAccess, FileOpenOptions OpenOptions, bool inherit)
    {
      AttributeFlags flags = AttributeFlags.CaseInsensitive;
      if (inherit)
        flags |= AttributeFlags.Inherit;
      using (ObjectAttributes obja = new ObjectAttributes(name, flags))
      {
        IntPtr handle;
        IoStatus iostatus = new IoStatus();
        int status = NtOpenFile(out handle, DesiredAccess, obja, iostatus, ShareAccess, OpenOptions);
        StatusToNtException(status);
        return new SafeFileHandle(handle, true);
      }
    }

    [StructLayout(LayoutKind.Sequential)]
    public struct FILETIME
    {
      public uint DateTimeLow;
      public uint DateTimeHigh;
    }

    [StructLayout(LayoutKind.Sequential)]
    struct ByHandleFileInformation
    {
      public uint FileAttributes;
      public FILETIME CreationTime;
      public FILETIME LastAccessTime;
      public FILETIME LastWriteTime;
      public uint VolumeSerialNumber;
      public uint FileSizeHigh;
      public uint FileSizeLow;
      public uint NumberOfLinks;
      public uint FileIndexHigh;
      public uint FileIndexLow;
    }

    [DllImport("kernel32.dll", SetLastError = true)]
    static extern bool GetFileInformationByHandle(
       SafeFileHandle hFile,
       out ByHandleFileInformation lpFileInformation
    );

    public static string GetFileId(string path)
    {
      using (SafeFileHandle file = OpenFile(path, FileAccessRights.MaximumAllowed, ShareMode.None, FileOpenOptions.None, false))
      {
        ByHandleFileInformation info;
        if (!GetFileInformationByHandle(file, out info))
        {
          throw new Win32Exception(Marshal.GetLastWin32Error());
        }
        ulong file_id = (ulong)info.FileIndexHigh << 32 | info.FileIndexLow;
        return Encoding.Unicode.GetString(BitConverter.GetBytes(file_id));
      }
    }

    public static SafeFileHandle OpenFileById(SafeFileHandle volume_handle, string id, FileAccessRights DesiredAccess, ShareMode ShareAccess, FileOpenOptions OpenOptions, bool inherit)
    {
      AttributeFlags flags = AttributeFlags.CaseInsensitive;
      if (inherit)
        flags |= AttributeFlags.Inherit;
      StringBuilder name_builder = new StringBuilder();
      using (ObjectAttributes obja = new ObjectAttributes(id, flags, volume_handle.DangerousGetHandle(), null, null))
      {
        IntPtr handle;
        IoStatus iostatus = new IoStatus();
        int status = NtOpenFile(out handle, DesiredAccess, obja, iostatus, ShareAccess, OpenOptions | FileOpenOptions.OpenByFileId);
        StatusToNtException(status);
        return new SafeFileHandle(handle, true);
      }
    }

    [DllImport("ntdll.dll")]
    public static extern int NtCreateProcess(out IntPtr ProcessHandle, ProcessAccessRights DesiredAccess, ObjectAttributes ObjectAttributes, IntPtr InheritFromProcessHandle, [MarshalAs(UnmanagedType.U1)] bool InheritHandles, IntPtr SectionHandle, IntPtr DebugPort, IntPtr ExceptionPort);
    [DllImport("ntdll.dll")]
    public static extern int NtCreateProcessEx(out IntPtr ProcessHandle, ProcessAccessRights DesiredAccess, ObjectAttributes ObjectAttributes, IntPtr InheritFromProcessHandle, ProcessCreateFlags Flags, IntPtr SectionHandle, IntPtr DebugPort, IntPtr ExceptionPort, int Unused);


    public static SafeKernelObjectHandle CreateProcessEx(SafeKernelObjectHandle ParentProcess, ProcessCreateFlags Flags, SafeKernelObjectHandle SectionHandle)
    {
      IntPtr process;
      IntPtr parent_process = ParentProcess != null ? ParentProcess.DangerousGetHandle() : CurrentProcess;
      IntPtr section = SectionHandle != null ? SectionHandle.DangerousGetHandle() : IntPtr.Zero;
      StatusToNtException(NtCreateProcessEx(out process, ProcessAccessRights.MaximumAllowed, new ObjectAttributes(), parent_process, Flags, section, IntPtr.Zero, IntPtr.Zero, 0));
      return new SafeKernelObjectHandle(process, true);
    }

    public static SafeKernelObjectHandle CreateProcessEx(SafeKernelObjectHandle SectionHandle)
    {
      return CreateProcessEx(null, ProcessCreateFlags.None, SectionHandle);
    }

    [DllImport("ntdll.dll")]
    public static extern int RtlCreateProcessParametersEx(
            [Out] out IntPtr pProcessParameters,
            [In] UnicodeString ImagePathName,
            [In] UnicodeString DllPath,
            [In] UnicodeString CurrentDirectory,
            [In] UnicodeString CommandLine,
            [In, MarshalAs(UnmanagedType.LPArray)] byte[] Environment,
            [In] UnicodeString WindowTitle,
            [In] UnicodeString DesktopInfo,
            [In] UnicodeString ShellInfo,
            [In] UnicodeString RuntimeData,
            uint Flags);

    [DllImport("ntdll.dll")]
    public static extern int NtOpenProcess(
      out IntPtr ProcessHandle,
      ProcessAccessRights DesiredAccess,
      [In] ObjectAttributes ObjectAttributes,
      [In] ClientId ClientId
    );

    public static SafeKernelObjectHandle OpenProcess(int pid, ProcessAccessRights desired_access)
    {
      IntPtr process;
      ClientId client_id = new ClientId();
      client_id.UniqueProcess = new IntPtr(pid);
      StatusToNtException(NtOpenProcess(out process, desired_access, new ObjectAttributes(), client_id));
      return new SafeKernelObjectHandle(process, true);
    }

    private static UnicodeString GetString(string s)
    {
      return s != null ? new UnicodeString(s) : null;
    }

    public static IntPtr CreateProcessParameters(
            string ImagePathName,
            string DllPath,
            string CurrentDirectory,
            string CommandLine,
            byte[] Environment,
            string WindowTitle,
            string DesktopInfo,
            string ShellInfo,
            string RuntimeData,
            uint Flags)
    {
      IntPtr ret;

      int status = RtlCreateProcessParametersEx(out ret, GetString(ImagePathName), GetString(DllPath), GetString(CurrentDirectory),
        GetString(CommandLine), Environment, GetString(WindowTitle), GetString(DesktopInfo), GetString(ShellInfo), GetString(RuntimeData), Flags);
      StatusToNtException(status);

      return ret;
    }

    [DllImport("ntdll.dll")]
    public static extern void RtlDestroyProcessParameters(IntPtr pProcessParameters);

    public const uint MaximumAllowed = 0x02000000;

    [DllImport("ntdll.dll")]
    public static extern int NtCreateUserProcess(
      out IntPtr ProcessHandle, out IntPtr ThreadHandle,
      uint ProcessDesiredAccess,
      uint ThreadDesiredAccess,
      ObjectAttributes ProcessObjectAttributes,
      ObjectAttributes ThreadObjectAttributes,
      ProcessCreateFlags ProcessFlags,
      ThreadCreateFlags ThreadFlags,
      IntPtr ProcessParameters,
      [In, Out] ProcessCreateInfo CreateInfo,
      [In, Out] ProcessAttributeList AttributeList
    );

    [DllImport("ntdll.dll")]
    public static extern int NtTerminateProcess(IntPtr ProcessHandle, int ExitCode);
    [DllImport("ntdll.dll")]
    public static extern int NtResumeThread(IntPtr ThreadHandle, out int PreviousSuspendCount);

    [DllImport("ntdll.dll")]
    public static extern int NtCreateSection(out IntPtr SectionHandle, uint DesiredAccess,
        [Out] ObjectAttributes ObjectAttributes, [Out] LargeInteger SectionSize,
        ProtectionType Protect, SectionAttributes Attributes,
        IntPtr FileHandle);

    public static SafeKernelObjectHandle CreateImageSection(SafeFileHandle FileHandle)
    {
      IntPtr section_handle;
      ObjectAttributes obj_attr = new ObjectAttributes();
      StatusToNtException(NtCreateSection(out section_handle, MaximumAllowed, obj_attr, null, ProtectionType.Execute, SectionAttributes.Image, FileHandle.DangerousGetHandle()));

      return new SafeKernelObjectHandle(section_handle, true);
    }

    public enum FileInformationClass
    {
      FileDirectoryInformation = 1,
      FileFullDirectoryInformation,
      FileBothDirectoryInformation,
      FileBasicInformation,
      FileStandardInformation,
      FileInternalInformation,
      FileEaInformation,
      FileAccessInformation,
      FileNameInformation,
      FileRenameInformation,
      FileLinkInformation,
      FileNamesInformation,
      FileDispositionInformation,
      FilePositionInformation,
      FileFullEaInformation,
      FileModeInformation,
      FileAlignmentInformation,
      FileAllInformation,
      FileAllocationInformation,
      FileEndOfFileInformation,
      FileAlternateNameInformation,
      FileStreamInformation,
      FilePipeInformation,
      FilePipeLocalInformation,
      FilePipeRemoteInformation,
      FileMailslotQueryInformation,
      FileMailslotSetInformation,
      FileCompressionInformation,
      FileObjectIdInformation,
      FileCompletionInformation,
      FileMoveClusterInformation,
      FileQuotaInformation,
      FileReparsePointInformation,
      FileNetworkOpenInformation,
      FileAttributeTagInformation,
      FileTrackingInformation,
      FileIdBothDirectoryInformation,
      FileIdFullDirectoryInformation,
      FileValidDataLengthInformation,
      FileShortNameInformation,
      FileIoCompletionNotificationInformation,
      FileIoStatusBlockRangeInformation,
      FileIoPriorityHintInformation,
      FileSfioReserveInformation,
      FileSfioVolumeInformation,
      FileHardLinkInformation,
      FileProcessIdsUsingFileInformation,
      FileNormalizedNameInformation,
      FileNetworkPhysicalNameInformation,
      FileIdGlobalTxDirectoryInformation,
      FileIsRemoteDeviceInformation,
      FileUnusedInformation,
      FileNumaNodeInformation,
      FileStandardLinkInformation,
      FileRemoteProtocolInformation,
      FileRenameInformationBypassAccessCheck,
      FileLinkInformationBypassAccessCheck,
      FileVolumeNameInformation,
      FileIdInformation,
      FileIdExtdDirectoryInformation,
      FileReplaceCompletionInformation,
      FileHardLinkFullIdInformation,
      FileIdExtdBothDirectoryInformation,
      FileMaximumInformation
    }

    [DllImport("ntdll.dll")]
    public static extern int NtSetInformationFile(
      SafeFileHandle FileHandle,
      [Out] IoStatus IoStatusBlock,
      byte[] FileInformation,
      int Length,
      FileInformationClass FileInformationClass
    );

    private static byte[] StructToBytes<T>(T value)
    {
      int length = Marshal.SizeOf<T>();
      byte[] ret = new byte[length];
      IntPtr buffer = Marshal.AllocHGlobal(length);
      try
      {
        Marshal.StructureToPtr(value, buffer, false);
        Marshal.Copy(buffer, ret, 0, ret.Length);
      }
      finally
      {
        if (buffer != IntPtr.Zero)
        {
          Marshal.FreeHGlobal(buffer);
        }
      }
      return ret;
    }

    public static void DeleteFileByHandle(SafeFileHandle handle)
    {
      IoStatus iostatus = new IoStatus();
      FileDispositionInformation deletefile = new FileDispositionInformation() { DeleteFile = true };
      byte[] buffer = StructToBytes(deletefile);
      StatusToNtException(NtSetInformationFile(handle, iostatus, buffer,
              buffer.Length, FileInformationClass.FileDispositionInformation));
    }

    public static void CreateHardlink(SafeFileHandle handle, string linkname)
    {
      IoStatus iostatus = new IoStatus();
      FileLinkRenameInformation link = new FileLinkRenameInformation();
      link.ReplaceIfExists = true;
      link.RootDirectory = IntPtr.Zero;
      link.FileName = linkname;
      link.FileNameLength = linkname.Length * 2;
      byte[] buffer = StructToBytes(link);
      StatusToNtException(NtSetInformationFile(handle, iostatus, buffer,
              buffer.Length, FileInformationClass.FileLinkInformation));
    }

    public static void CreateHardlink(string file, string linkname)
    {
      using (SafeFileHandle handle = OpenFile(file, FileAccessRights.MaximumAllowed, ShareMode.Read, FileOpenOptions.NonDirectoryFile, false))
      {
        CreateHardlink(handle, linkname);
      }
    }

    [Flags]
    public enum QueryFullProcessImageNameFlags
    {
      Win32Path = 0,
      NativePath = 1,
    }

    [DllImport("kernel32.dll", CharSet = CharSet.Unicode, SetLastError = true)]
    public static extern bool QueryFullProcessImageName(
      SafeKernelObjectHandle hProcess,
      QueryFullProcessImageNameFlags dwFlags,
      StringBuilder lpExeName,
      ref int lpdwSize);


    [DllImport("ntdll.dll")]
    public static extern int NtQueryInformationProcess(SafeKernelObjectHandle ProcessHandle,
      ProcessInfoClass ProcessInformationClass,
      IntPtr ProcessInformation,
      int ProcessInformationLength,
      [Out] out int ReturnLength
      );

    public struct SidAndAttributes
    {
      public IntPtr Sid;
      public int Attributes;
    };

    [DllImport("ntdll.dll", CharSet = CharSet.Unicode)]
    public static extern int NtCreateLowBoxToken(
      out IntPtr token,
      IntPtr original_token,
      GenericAccessRights access,
      ObjectAttributes object_attribute,
      byte[] appcontainer_sid,
      int capabilityCount,
      SidAndAttributes[] capabilities,
      int handle_count,
      IntPtr[] handles);

    public static SafeKernelObjectHandle CreateLowBoxToken(SafeKernelObjectHandle original_token,
      string package_sid, params SafeHandle[] handles)
    {
      IntPtr token;
      SecurityIdentifier sid = new SecurityIdentifier(package_sid);
      byte[] sid_bin = new byte[sid.BinaryLength];
      sid.GetBinaryForm(sid_bin, 0);

      IntPtr[] handle_array = null;
      int handle_count = handles != null ? handles.Length : 0;
      if (handle_count > 0)
      {
        handle_array = new IntPtr[handle_count];
        for (int i = 0; i < handle_count; ++i)
        {
          handle_array[i] = handles[i].DangerousGetHandle();
        }
      }

      StatusToNtException(NtCreateLowBoxToken(out token, original_token.DangerousGetHandle(), GenericAccessRights.MaximumAllowed,
        new ObjectAttributes(), sid_bin, 0, null, handle_count, handle_array));

      return new SafeKernelObjectHandle(token, true);
    }

    [DllImport("ntdll.dll", CharSet = CharSet.Unicode)]
    public static extern int NtOpenProcessTokenEx(
      IntPtr ProcessHandle,
      GenericAccessRights DesiredAccess,
      AttributeFlags HandleAttributes,
      out IntPtr TokenHandle);

    [DllImport("ntdll.dll")]
    public static extern int NtOpenThreadTokenEx(
      IntPtr ThreadHandle,
      GenericAccessRights DesiredAccess,
      [MarshalAs(UnmanagedType.U1)] bool OpenAsSelf,
      AttributeFlags HandleAttributes,
      out IntPtr TokenHandle
    );

    public static SafeKernelObjectHandle OpenProcessToken(SafeKernelObjectHandle process, bool duplicate)
    {
      IntPtr new_token;
      StatusToNtException(NtOpenProcessTokenEx(process.IsInvalid ? new IntPtr(-1) : process.DangerousGetHandle(),
        GenericAccessRights.MaximumAllowed, AttributeFlags.None, out new_token));
      SafeKernelObjectHandle ret = new SafeKernelObjectHandle(new_token, true);
      if (duplicate)
      {
        try
        {
          return NtSystemCalls.DuplicateToken(ret, TokenType.Primary);
        }
        finally
        {
          ret.Close();
        }
      }
      return ret;
    }

    public static SafeKernelObjectHandle OpenProcessToken(SafeKernelObjectHandle process)
    {
      return OpenProcessToken(process, false);
    }

    public static SafeKernelObjectHandle OpenProcessToken()
    {
      return OpenProcessToken(false);
    }

    public static SafeKernelObjectHandle OpenProcessToken(bool duplicate)
    {
      return OpenProcessToken(new SafeKernelObjectHandle(new IntPtr(-1), false), duplicate);
    }

    [DllImport("ntdll.dll", CharSet = CharSet.Unicode)]
    public static extern int NtDuplicateToken(
        IntPtr ExistingTokenHandle,
        GenericAccessRights DesiredAccess,
        ObjectAttributes ObjectAttributes,
        bool EffectiveOnly,
        TokenType TokenType,
        out IntPtr NewTokenHandle
    );

    public static SafeKernelObjectHandle DuplicateToken(SafeKernelObjectHandle existing_token, TokenType type, SecurityImpersonationLevel level)
    {
      IntPtr new_token;
      SecurityQualityOfService sqos = null;
      if (type == TokenType.Impersonation)
      {
        sqos = new SecurityQualityOfService();
        sqos.ImpersonationLevel = level;
        sqos.ContextTrackingMode = SecurityContextTrackingMode.Static;
      }

      using (ObjectAttributes obja = new ObjectAttributes(null, AttributeFlags.None, IntPtr.Zero, sqos, null))
      {
        StatusToNtException(NtDuplicateToken(existing_token.DangerousGetHandle(),
          GenericAccessRights.MaximumAllowed, obja, false, type, out new_token));
        return new SafeKernelObjectHandle(new_token, true);
      }
    }

    public static SafeKernelObjectHandle DuplicateToken(SafeKernelObjectHandle existing_token, TokenType type)
    {
      return DuplicateToken(existing_token, type, SecurityImpersonationLevel.Impersonation);
    }

    public static IntPtr CurrentProcess { get { return new IntPtr(-1); } }
    public static IntPtr CurrentThread { get { return new IntPtr(-2); } }

    [DllImport("ntdll.dll")]
    public static extern int NtClose(IntPtr handle);

    [DllImport("ntdll.dll")]
    public static extern int NtDuplicateObject(
      IntPtr SourceProcessHandle,
      IntPtr SourceHandle,
      IntPtr TargetProcessHandle,
      out IntPtr TargetHandle,
      GenericAccessRights DesiredAccess,
      AttributeFlags HandleAttributes,
      DuplicateObjectOptions Options
    );

    private static IntPtr DuplicateObject(IntPtr handle)
    {
      IntPtr new_handle;

      StatusToNtException(NtDuplicateObject(CurrentProcess, handle, CurrentProcess, out new_handle,
        GenericAccessRights.None, AttributeFlags.None, DuplicateObjectOptions.SameAccess | DuplicateObjectOptions.SameAttributes));

      return new_handle;
    }

    public static SafeKernelObjectHandle DuplicateObject(SafeKernelObjectHandle handle)
    {
      return new SafeKernelObjectHandle(DuplicateObject(handle.DangerousGetHandle()), true);
    }

    public static SafeFileHandle DuplicateObject(SafeFileHandle handle)
    {
      return new SafeFileHandle(DuplicateObject(handle.DangerousGetHandle()), true);
    }

    public static SafeKernelObjectHandle DuplicateObject(SafeHandle handle)
    {
      return DuplicateObject(new SafeKernelObjectHandle(handle.DangerousGetHandle(), false));
    }

    [DllImport("ntdll.dll")]
    public static extern int NtQuerySystemInformation(
      SystemInformationClass SystemInformationClass,
      IntPtr SystemInformation,
      int SystemInformationLength,
      out int ReturnLength
    );

    public static IEnumerable<SystemHandleTableInfoEntry> GetHandles(int pid)
    {
      SafeHGlobalBuffer handleInfo = new SafeHGlobalBuffer(0x10000);
      try
      {
        int status = 0;
        int return_length = 0;
        while ((uint)(status = NtQuerySystemInformation(SystemInformationClass.SystemHandleInformation,
                                                 handleInfo.DangerousGetHandle(),
                                                 handleInfo.Length,
                                                 out return_length)) == 0xc0000004)
        {
          int length = handleInfo.Length * 2;
          handleInfo.Close();
          handleInfo = new SafeHGlobalBuffer(length);
        }
        StatusToNtException(status);

        IntPtr handleInfoBuf = handleInfo.DangerousGetHandle();
        int handle_count = Marshal.ReadInt32(handleInfoBuf);
        List<SystemHandleTableInfoEntry> ret = new List<SystemHandleTableInfoEntry>();
        handleInfoBuf += IntPtr.Size;
        for (int i = 0; i < handle_count; ++i)
        {
          SystemHandleTableInfoEntry entry = (SystemHandleTableInfoEntry)Marshal.PtrToStructure(handleInfoBuf, typeof(SystemHandleTableInfoEntry));

          if (pid == -1 || entry.UniqueProcessId == pid)
          {
            ret.Add(entry);
          }
          handleInfoBuf += Marshal.SizeOf(typeof(SystemHandleTableInfoEntry));
        }
        return ret;
      }
      finally
      {
        handleInfo.Close();
      }
    }

    public static IEnumerable<SystemHandleTableInfoEntry> GetHandles()
    {
      return GetHandles(-1);
    }

    [DllImport("ntdll.dll")]
    public static extern int NtGetNextProcess(
      IntPtr ProcessHandle,
      ProcessAccessRights DesiredAccess,
      AttributeFlags HandleAttributes,
      int Flags,
      out IntPtr NewProcessHandle
    );

    public static SafeKernelObjectHandle[] GetProcesses(ProcessAccessRights DesiredAccess)
    {
      List<SafeKernelObjectHandle> handles = new List<SafeKernelObjectHandle>();
      IntPtr current_handle = IntPtr.Zero;
      int status = NtGetNextProcess(current_handle, DesiredAccess, AttributeFlags.None, 0, out current_handle);
      while (status == 0)
      {
        handles.Add(new SafeKernelObjectHandle(current_handle, true));
        status = NtGetNextProcess(current_handle, DesiredAccess, AttributeFlags.None, 0, out current_handle);
      }
      return handles.ToArray();
    }

    public static int GetProcessSessionId(SafeKernelObjectHandle process)
    {
      using (SafeStructureOutBuffer<ProcessSessionInformation> session_info = new SafeStructureOutBuffer<ProcessSessionInformation>())
      {
        int return_length = 0;
        StatusToNtException(NtQueryInformationProcess(process, ProcessInfoClass.ProcessSessionInformation,
          session_info.DangerousGetHandle(), session_info.Length, out return_length));
        return session_info.Result.SessionId;
      }
    }

    public static int GetProcessId(SafeKernelObjectHandle process)
    {
      using (SafeStructureOutBuffer<ProcessBasicInformation> basic_info = new SafeStructureOutBuffer<ProcessBasicInformation>())
      {
        int return_length = 0;
        StatusToNtException(NtQueryInformationProcess(process, ProcessInfoClass.ProcessBasicInformation,
          basic_info.DangerousGetHandle(), basic_info.Length, out return_length));
        return basic_info.Result.UniqueProcessId.ToInt32();
      }
    }

    public static int GetProcessExitStatus(SafeKernelObjectHandle process)
    {
      using (SafeStructureOutBuffer<ProcessBasicInformation> basic_info = new SafeStructureOutBuffer<ProcessBasicInformation>())
      {
        int return_length = 0;
        StatusToNtException(NtQueryInformationProcess(process, ProcessInfoClass.ProcessBasicInformation,
          basic_info.DangerousGetHandle(), basic_info.Length, out return_length));
        return basic_info.Result.ExitStatus;
      }
    }


    public static string GetProcessImageName(SafeKernelObjectHandle process, bool native_path)
    {
      StringBuilder builder = new StringBuilder(256);
      int size = 256;
      if (!NtSystemCalls.QueryFullProcessImageName(process, native_path ?
        NtSystemCalls.QueryFullProcessImageNameFlags.NativePath : NtSystemCalls.QueryFullProcessImageNameFlags.Win32Path,
          builder, ref size))
      {
        throw new Win32Exception(Marshal.GetLastWin32Error());
      }
      return builder.ToString();
    }

    [Flags]
    public enum FinalPathNameFlags
    {
      None = 0,
      NameGuid = 1,
      NameNt = 2,
      NameNone = 4,
      Opened = 8,
    }

    [DllImport("kernel32.dll", CharSet = CharSet.Unicode, SetLastError = true)]
    public static extern int GetFinalPathNameByHandle(SafeFileHandle hFile, StringBuilder lpszFilePath, int cchFilePath, FinalPathNameFlags dwFlags);

    public static string GetPathName(SafeFileHandle handle, FinalPathNameFlags flags)
    {
      StringBuilder builder = new StringBuilder(1000);
      if (GetFinalPathNameByHandle(handle, builder, builder.Capacity, flags) == 0)
      {
        throw new Win32Exception(Marshal.GetLastWin32Error());
      }
      return builder.ToString();
    }


    [DllImport("ntdll.dll")]
    public static extern int NtSetInformationToken(
      SafeKernelObjectHandle TokenHandle,
      TokenInformationClass TokenInformationClass,
      IntPtr TokenInformation,
      int TokenInformationLength);

    [DllImport("ntdll.dll")]
    public static extern int NtQueryInformationToken(
      SafeKernelObjectHandle TokenHandle,
      TokenInformationClass TokenInformationClass,
      IntPtr TokenInformation,
      int TokenInformationLength,
      out int ReturnLength);

    public static void SetTokenSessionId(SafeKernelObjectHandle token, int session_id)
    {
      SafeStructureBuffer buffer = new SafeStructureBuffer(session_id);
      StatusToNtException(NtSetInformationToken(token, TokenInformationClass.TokenSessionId, buffer.DangerousGetHandle(), buffer.Length));
    }

    [StructLayout(LayoutKind.Sequential)]
    struct Luid
    {
      public uint LowPart;
      public int HighPart;
    }

    [StructLayout(LayoutKind.Sequential)]
    struct LuidAndAttributes
    {
      public Luid Luid;
      public int Attributes;
    }

    [StructLayout(LayoutKind.Sequential)]
    class TokenPrivileges
    {
      public int PrivilegeCount;
      public LuidAndAttributes Privilege;
    }

    [DllImport("advapi32.dll", CharSet=CharSet.Unicode, SetLastError= true)]
    [return: MarshalAs(UnmanagedType.Bool)]
    private static extern bool LookupPrivilegeValue(
      string lpSystemName,
      string lpName,
      out Luid lpLuid
    );

    [DllImport("advapi32.dll", CharSet=CharSet.Unicode,  SetLastError= true)]
    private static extern bool AdjustTokenPrivileges(
      SafeKernelObjectHandle TokenHandle,
      [MarshalAs(UnmanagedType.Bool)] bool DisableAllPrivileges,
      TokenPrivileges NewState,
      int BufferLength,
      IntPtr PreviousState,
      IntPtr ReturnLength
    );

    const int ErrorNotAllAssigned = 1300;
    const int SePrivilegeEnabled = 2;

    public const string SeDebugPrivilege = "SeDebugPrivilege";

    public static bool SetPrivilege(SafeKernelObjectHandle hToken, string lpszPrivilege, bool bEnablePrivilege)
    {
      TokenPrivileges tp = new TokenPrivileges();
      Luid luid;

      if (!LookupPrivilegeValue(null, lpszPrivilege, out luid))
      {
        throw new Win32Exception(Marshal.GetLastWin32Error());
      }

      tp.PrivilegeCount = 1;
      tp.Privilege.Luid = luid;
      if (bEnablePrivilege)
      {
        tp.Privilege.Attributes = SePrivilegeEnabled;
      }
      else
      {
        tp.Privilege.Attributes = 0;
      }

      if (!AdjustTokenPrivileges(hToken, false, tp, Marshal.SizeOf(tp), IntPtr.Zero, IntPtr.Zero))
      {
        throw new Win32Exception(Marshal.GetLastWin32Error());
      }

      if (Marshal.GetLastWin32Error() == ErrorNotAllAssigned)
      {
        return false;
      }

      return true;
    }

    public static bool EnableDebugPrivilege()
    {
      using (SafeKernelObjectHandle token = OpenProcessToken())
      {
        return SetPrivilege(token, SeDebugPrivilege, true);
      }
    }
  }
}
